!+
! Module: epreject_mods
!
! Authors:  Konrad Dennerl, Hermann Brunner, Vadim Burwitz - MPE
! Co-Author:-)  Michael Freyberg - MPE
!          
! Description: 
! 
! Comments: 
! 
! $Id: module.f90,v 1.5 1998/01/13 08:39:15 gvacanti Exp $
!- 
module epreject_mod

  use eventAttributes
  use errorHandling
  use oal
  use cal
  use param
  use dal
  
  implicit none

  private

  public epreject

contains

  !---------------------------------------------------------------------
  subroutine epreject
  !---------------------------------------------------------------------
  
   character(len=10), parameter :: CodeVersion = '5.9'
   real(kind=Single) :: hmed, expo, exp1k, h1, h2, bg, bgfac, offsig
   real(kind=Single) :: map_min, map_max
   real(kind=Single), dimension(0:0), target :: rdum
   integer(kind=Int16), dimension(0:0), target :: i2dum
   integer(kind=Int32) :: i, j, k, ix, iy, ky, jy, ixx, iyy, nevent, kval
   integer(kind=Int32) :: koff, ksel, iccd, ccd, idum, isubmode, flagged
   integer(kind=Int32) :: noffcorr, ios = 0, badX, noffsets, nbadpix
   integer(kind=Int32) :: winX0, winX1, winY0, winY1, win4Y0
   integer(kind=Int32) :: count3, pix3, count5, pix5, nsou
   integer(kind=Int16) :: i2off, i2pha
   integer(kind=Int32), parameter :: nxdim=64, nydim=200, nyStart=13
   integer(kind=Int32), parameter :: nccd=12, minspec=16, maxspec=65, &
                                     maxoffcorr=500 
   integer(kind=Int32), dimension(nccd,nydim,maxspec) :: i4spc
   integer(kind=Int32), dimension(nydim,maxspec) :: i4buf, i4buff
   integer(kind=Int32), dimension(nxdim,nydim) :: i4pmap, i30map
   integer(kind=Int32), dimension(nydim) :: i4arr1
   integer(kind=Int16), dimension(nxdim,nydim), target :: i2pmap
   real(kind=Single), dimension(nydim,maxspec) :: spcfrc
   real(kind=Single), dimension(0:12) :: noiseparam
   integer(kind=Int32), dimension(nydim) :: i4med, indx 
   real(kind=Single), dimension(:), pointer :: hlo, hhi
   integer(kind=Int16), dimension(:), pointer :: val, rrawx
   real(kind=double), dimension(:), pointer :: timeCol
   integer(kind=Int16), dimension(:), pointer :: rawXCol, rawYCol, phaCol, offCol
   integer(kind=Int16), dimension(:), pointer :: noiseCol, flareCol, offXCol, bpRawXCol, bpRawYCol, yExtCol 
   integer(kind=Int8), dimension(:), pointer :: ccdCol     
   integer(kind=Int16), dimension(:,:), pointer :: i2offset, i2map
   integer(kind=int32), dimension(2) :: npix
   real(kind=Single), dimension(:,:), pointer :: r4med
   real(kind=Single), dimension(:,:,:), pointer :: noisemap
   real(kind=Single), dimension(:,:), pointer :: MasterOffset
   logical, dimension(nxdim,nydim) :: mask = .false.
   logical :: exist, withoffsetmap, calibrated_noise, calibrated_offset
   character(len=1024) :: string, fileName, shortFileName, odifile
   character(len=32) :: submode
   logical :: ldebug = .true.

!-- eprej flare routine VB
   integer(kind=Int32), dimension(:), allocatable :: lc
   real(kind=single), dimension(:), allocatable :: lcs
   real(kind=double)   :: tbinsize,t0,tz
   real(kind=double)   :: ss
   integer(kind=Int32) :: kmax,ndlc,sbox
   integer(kind=Int32) :: npharangeelements
 ! integer(kind=Int32) :: nsmoothelements
   real(kind=double),   dimension(0:2) :: smo
   integer(kind=Int32), dimension(0:1) :: thrpha
   real(kind=double) :: thrsig

   type(DataSetT) :: infile, odiset, outfile
   type(arrayT)  :: odiarr, ima
   type(TableT)  :: events, offsets, badpix
   type(ColumnT) :: ccdnr, rawX, rawY, pha, offCor, noise, flare, offX
   type(ColumnT) :: bpRawX, bpRawY, yExtent 
   type(ColumnT) :: time

!-- eprej lower threshold handling MJF
   integer(kind=Int16) :: LowerThreshold, An_CMCOR, An_CMLOT
   integer(kind=Int32) :: Nbelow
   integer(kind=Int32), dimension(:), pointer :: flagCol
   type(ColumnT) :: flag

   write(errstr,*) 'EPREJECT ',trim(CodeVersion),' is running...'
   call message(AppMsg,Verbose)


   i4spc = 0
   i4buf = 0
   i4buff = 0
   i4pmap = 0
   i30map = 0
   i2pmap = 0
   spcfrc = 0.
!
! open the event file in LOW_MEMORY:
! because EXPOSURE extensions are untouched and rather large
!
   infile = dataSet(stringParameter("eventset"),MODIFY,LOW_MEMORY)
   if (.not. hasBlock(infile,"EVENTS")) then 
     call error("noEvents","No EVENTS extension in input data set")
   endif

   events = table(infile,"EVENTS",MODIFY)
   nevent = numberOfRows(events)

   if (hasAttribute(events,"SUBMODE")) then
     submode = stringAttribute(events,"SUBMODE")
   else
     call warning("WrongSubMode","Could not read SUBMODE keyword; FF mode assumed") 
     submode = "PrimeFullWindow"
   endif
   isubmode = modeId(trim(submode))

   if (.not.booleanParameter("withsoftflarescreening")) then

   offsig = realParameter("sigma")
   do i=0,12
     noiseparam(i) = realParameter("noiseparameters",i)
   end do

! open offset map
   withoffsetmap=booleanParameter("withoffsetmap")
   if (withoffsetmap) then
     shortfileName = trim(stringParameter('set'))
     call OAL_expandFileName(shortfileName,fileName)
     call OAL_setState(fileName)
     if (OAL_hasAssociatedSet(OFFSET_DATA)) then
       odifile=OAL_selectAssociatedSet(OFFSET_DATA)       
       odiset  = dataSET(odifile, READ)
       odiarr = array(odiset, "PNODI1", READ)
       i2offset => int16array2data(odiarr)
       write(errstr,'(2a)') 'Reading offset map ',trim(odifile)
       call message(AppMsg,Sparse)
     else

       withoffsetmap=.false.
       call warning("NoOffsetMap","No offset map found; using 20 ADU image instead")
       i2pmap = 0
       i2offset => i2pmap

     end if
   endif

   expo=Real32Attribute(events,"EXPOSURE")
   exp1k=expo/1000.  ! s -> ks
!! expo =expo / 2.16 ! eFF

   ccdnr = column(events,"CCDNR",READ)
   ccdCol      => int8Data(ccdnr)

   iccd=ccdCol(0)
   call CAL_setState(instrument=EPN,ccdModeId=isubmode,ccdChipId=iccd)

   map_min = -1.0
   map_max = -1.0
   call message(AppMsg,Noisy,'reading from CCF: pnNoiseMap')
   call CAL_pnNoiseMap(rrawx,noisemap)
   map_min = MINVAL(noisemap(:,:,:))
   map_max = MAXVAL(noisemap(:,:,:))
   if (map_max .le. 0.0 .and. map_min .le. 0.0) then
     calibrated_noise = .FALSE.
   else
     calibrated_noise = .TRUE.
   endif

   if (.not. calibrated_noise) then
     write(errstr,*) "no noise calibration for submode ",trim(SUBMODE), &
                     ", skipping noise flagging"
     call warning("notCalibrated")
   endif

   map_min = -1.0
   map_max = -1.0
   call message(AppMsg,Noisy,'reading from CCF: pnMedianMap')
   call CAL_pnMedianMap(r4med)
   map_min = MINVAL(r4med(:,:))
   map_max = MAXVAL(r4med(:,:))
   if (map_max .le. 0.0 .and. map_min .le. 0.0) then
     calibrated_offset = .FALSE.
     noffcorr = 1
     hlo => rdum
     hlo(0) = -2.2E-1
     hhi => rdum
     hhi(0) = +1.0E10
     val => i2dum
     val(0) = 0
   else
     calibrated_offset = .TRUE.
     call CAL_pnOffsetCorrValues(hlo,hhi,val)
     noffcorr = size(val)
   endif
   r4med=r4med*exp1k

   if (withoffsetmap) then
! get MasterOffsetMap        
     call message(AppMsg,Noisy,'reading from CCF: pnMasterOffset')
     call CAL_pnMasterOffset(MasterOffset)
   endif

   if (.not. withoffsetmap .and. .not. calibrated_offset) then
     write(errstr,*) "no offset calibration for submode ",trim(SUBMODE), &
                     ", skipping rest of the task"
     call warning("notCalibrated")
   else

   rawX = column(events,"RAWX",READ)
   rawY = column(events,"RAWY",READ)
   pha  = column(events,"PHA",MODIFY)
   flag = column(events,"FLAG",MODIFY)

   rawXCol     => int16Data(rawX)
   rawYCol     => int16Data(rawY)
   phaCol      => int16Data(pha)
   flagCol     => int32Data(flag)

   winX0 = int16Attribute(events,"WINDOWX0") 
   winY0 = int16Attribute(events,"WINDOWY0")     
   winX1 = int16Attribute(events,"WINDOWDX") + winX0 - 1
   winY1 = int16Attribute(events,"WINDOWDY") + winY0 - 1
   mask(winX0:winX1,winY0:winY1)=.true.

!
! read and check keywords relevant for lower threshold [MJF]
!
   if (hasAttribute(events,"AN_CMLOT")) then
     An_CMLOT = int16Attribute(events,"AN_CMLOT")
   else
     call warning("MissingHKvalue","Could not read AN_CMLOT keyword; assuming 532 [adu]") 
     An_CMLOT = 532
   endif
   if (hasAttribute(events,"AN_CMCOR")) then
     An_CMCOR = int16Attribute(events,"AN_CMCOR")
   else
     call warning("MissingHKvalue","Could not read AN_CMCOR keyword; assuming 512 [adu]") 
     An_CMCOR = 512
   endif
   LowerThreshold = An_CMLOT - An_CMCOR
   if ((LowerThreshold .lt. 16) .or. (LowerThreshold .gt. 100)) then
     call warning("UnusualHKvalue","Unusual lower threshold, check keywords AN_CMLOT and AN_CMCOR.") 
   endif


! append OFF_COR column
   if (hasColumn(events,"OFF_COR")) then
     offCor = column(events,"OFF_COR",MODIFY)  
     offCol => int16Data(offCor)
   else
     offCor = addColumn (events, "OFF_COR", Integer16, "")
     offCol => int16Data(offCor)
     offCol = 0
   end if   

!-----------------------------
! bin 20 & 30-40 ADU images    
   do i=0,nevent-1
     if (iccd /= ccdCol(i)) call error("Single CCD event file required") 
     if ((phaCol(i)+offCol(i))==20) then
       i4pmap(rawXCol(i),rawYCol(i))=i4pmap(rawXCol(i),rawYCol(i))+1
     end if
     if (phaCol(i)+offCol(i)>=30 .and. phaCol(i)+offCol(i)<=40) then
       i30map(rawXCol(i),rawYCol(i))=i30map(rawXCol(i),rawYCol(i))+1
     end if
   end do

! read in bad columns and remove from mask
   string = stringParameter("badcolumnset")   
   inquire(file=string,exist=exist)
   if (exist) then
     open(1,file=string,status="old")
     ios=0
     do while (ios == 0)
       read (1,*,iostat=ios) ccd, badX
       if (ios /= 0) exit
       if (ccd==iccd) mask(badX,winY0:winY1)=.false.
     end do       
     close(1)
   else
     if (.not. withoffsetmap) call warning ("No bad column table found","")
   end if

! get offset columns and remove from mask
   noffsets=0
   if (hasBlock(infile,"OFFSETS")) then
     offsets = table(infile,"OFFSETS",READ)
     noffsets = numberOfRows(offsets)

     if (noffsets > 0) then
       offX = column(offsets,"RAWX",READ)
       offXCol => int16Data(offX)

       do i=1,noffsets
         badX=offXCol(i-1)
         mask(badX,winY0:winY1)=.false.
       end do

       call release(offX)
     end if

     call release(offsets)   
   end if

! get badpixels and remove from mask
   nbadpix=0
   if (hasBlock(infile,"BADPIX")) then
     badpix = table(infile,"BADPIX",READ)
     nbadpix = numberOfRows(badpix)
   
     if (nbadpix > 0) then
       bpRawX = column(badpix,"RAWX",READ) 
       bpRawY = column(badpix,"RAWY",READ)
       yExtent = column(badpix,"YEXTENT",READ)

       bpRawXCol => int16Data(bpRawX)
       bpRawYCol => int16Data(bpRawY)
       yExtCol => int16Data(yExtent)

       do i=1,nbadpix
         do iy=bpRawYCol(i-1),bpRawYCol(i-1)+yExtCol(i-1)-1
           mask(bpRawXCol(i-1),iy)=.false.
         end do
       end do 

       call release(bpRawX)
       call release(bpRawY)
       call release(yExtent)
     end if

     call release(badpix)
   end if

! remove rows < 16 from mask
  mask(:,1:15)=.false.
   
! source search in 30-40 ADU map
   nsou = 0
   do ix=winX0,winX1
     do iy=winY0,winY1 
       count3=0
       pix3=0
       do ixx=ix-1,ix+1
         do iyy=iy-1,iy+1    
           if (ixx >= 1 .and. ixx <= nxdim .and. & 
             iyy >= 1 .and. iyy <= nydim) then
             if (mask(ixx,iyy)) then     
               count3=count3+i30map(ixx,iyy) 
               pix3=pix3+1
             end if
           end if   
         end do
       end do
       count5=0
       pix5=0 
       do ixx=ix-3,ix+3
         do iyy=iy-3,iy+3    
           if (ixx >= 1 .and. ixx <= nxdim .and. & 
             iyy >= 1 .and. iyy <= nydim) then
             if (mask(ixx,iyy)) then     
               count5=count5+i30map(ixx,iyy) 
               pix5=pix5+1
             end if
           end if   
         end do
       end do
       count5=count5-count3
       pix5=pix5-pix3
       if (pix3>0 .and. pix5>0) then
         bgfac=real(pix3)/real(pix5)
         bg=real(count5)*bgfac 
         if (count3>2 .and. count3-bg>10.*sqrt(bg)) then
           nsou = nsou+1          
           write(errstr,'(a,2i4)') &
           "Masking source at position: ", ix, iy
           call message(AppMsg,sparse)
           do ixx=ix-1,ix+1
             do iyy=iy-1,iy+1
               if (ixx >= 1 .and. ixx <= nxdim .and. & 
                   iyy >= 1 .and. iyy <= nydim) mask(ixx,iyy)=.false.
             end do
           end do  
         end if
       end if
     end do
   end do

   win4Y0 = (winY0-1)/4
   win4Y0 = 4*win4Y0+1

! search for additional hot pixels
!   do ix = winX0, winX1
!     do iy = win4Y0, winY1, 4
!       h4mean(1)=(i4pmap(ix,iy+1)+i4pmap(ix,iy+2)+i4pmap(ix,iy+3))/3.
!       h4mean(2)=(i4pmap(ix,iy)  +i4pmap(ix,iy+2)+i4pmap(ix,iy+3))/3.
!       h4mean(3)=(i4pmap(ix,iy)  +i4pmap(ix,iy+1)+i4pmap(ix,iy+3))/3.
!       h4mean(4)=(i4pmap(ix,iy)  +i4pmap(ix,iy+1)+i4pmap(ix,iy+2))/3.
!       do i=1,4
!         if (i4pmap(ix,iy+i-1)-h4mean(i)>20.*sqrt(h4mean(i))) mask(ix,iy+i-1)=.false.
!       end do
!     end do
!   end do 

   if (.not.withoffsetmap) then
     
    if (LowerThreshold == 20) then

     write(errstr,'(a,i2)') &
       "Determining offset corrections from 20 adu image for chip ", iccd
     call message(AppMsg,sparse)

     do ix = winX0, winX1
       do iy = win4Y0, winY1, 4

         kval=0
         hmed=0. 
         do i=0,3
           ky=iy+i
           if (mask(ix,ky)) then
             kval = kval + i4pmap(ix,ky) 
             hmed=hmed + r4med(ix-1,ky-1) 
           end if 
         end do

         if (abs(kval-hmed) > offsig*sqrt(real(kval))) then

           do ky=iy,iy+3 

             if (mask(ix,ky)) then
               kval = i4pmap(ix,ky) 
               hmed = r4med(ix-1,ky-1) 

               do i=0,noffcorr-1  
                 if (kval-hmed > hlo(i)*exp1k .and. &
                    kval-hmed <= hhi(i)*exp1k) &
                    i2pmap(ix,ky) = val(i)
              end do
             end if
           end do 
         end if         

       end do
     end do
  
    else
      call warning("Lower threshold != 20 adu: can't perform offset correction !")
    end if 

   else

     write(errstr,'(a,i2)') &
       "Determining offset corrections from offset map for chip ",iccd   
     call message(AppMsg,sparse)

! compute and subtract the median value for each column:
!  print *, winx0, winx1, winy0, winy1
     i2pmap=0
     do ix = winX0, winX1
       k = 0
       do iy = max(winY0,nyStart), winY1
         k = k + 1
         i4arr1(k) = i2offset(ix-1,iy-1)
       enddo
       call iindexx(k,i4arr1,indx)
       k = i4arr1(indx(k/2))
       !print *, ix, k
       do iy = winY0, winY1
         i2pmap(ix,iy) = i2offset(ix-1,iy-1) - k
       enddo
     enddo

! compute and subtract the median value for each row:
     do iy = max(winY0,nyStart), winY1
       k = 0
       do ix = winX0, winX1
       k = k + 1
         i4arr1(k) = i2pmap(ix,iy)
       enddo
       call iindexx(k,i4arr1,indx)
       k = i4arr1(indx(k/2))
       !print *, iy, k
       do ix = winX0, winX1
         i2pmap(ix,iy) = i2pmap(ix,iy) - k
       enddo
     enddo

     do ix=winX0,winX1
       do iy=winY0,winY1
         if (abs(i2pmap(ix,iy))>12) i2pmap(ix,iy)=0
       enddo
     enddo 
     i2pmap = -i2pmap

     do ix=winX0,winX1
       do iy=winY0,winY1
         i2pmap(ix,iy) = i2pmap(ix,iy)+MasterOffset(ix-1,iy-1)
!         if (MasterOffset(ix-1,iy-1) /= 0) print *, "TEST:", ix,iy,MasterOffset(ix-1,iy-1)

       enddo
     enddo 

   end if

!   i2pmap(:,1:20) = 0

   Nbelow = 0
   koff = 0
   do i=0,nevent-1
     ix = rawXCol(i)
     iy = rawYCol(i)
     if (min(ix,iy).lt.1 .or. ix.gt.nxdim .or. iy.gt.nydim) then
       call warning ("Invalid RawX, RawY")
       offCol(i) = 0  
     else 
       i2off = i2pmap(ix,iy)
       i2pha = phaCol(i) + offCol(i) - i2off       
       offCol(i) = i2off
       phaCol(i) = i2pha
       if (i2off /= 0) koff = koff + 1
       if (i2pha > 0 .and. i2pha <= maxspec) &
                           i4buf(iy,i2pha) = i4buf(iy,i2pha) + 1
     end if

     if ( phaCol(i) .lt. LowerThreshold ) then
!       flagCol(i) = ior(flagCol(i), EVATT_BELOW_ACCEPTANCE)
       flagCol(i) = ior(flagCol(i), EVATT_TRAILING)
       Nbelow = Nbelow + 1
     endif

   end do

   write(errstr,'(i9,a,i9,a)') koff, ' of', nevent, ' events selected for offset correction'
   call message(AppMsg,Sparse)
   write(errstr,'(i9,a,i5,a)') Nbelow,' events shifted below threshold of ',LowerThreshold,' [adu]'
   call message(AppMsg,Sparse)

   if (calibrated_noise) then

! append NOISE column
   if (hasColumn(events,"NOISE")) then
     noise = column(events,"NOISE",MODIFY)  
   else
     noise = addColumn (events, "NOISE", Integer16, "")
   end if   
   noiseCol => int16Data(noise)
   noiseCol = 0

! spatial smoothing/filtering of the spectra:
   do k = 1, maxspec
     do j = 1, 19
       i4buff(j,k) = i4buf(j,k)
     end do
     do j = 20, nydim
       ky = 0
       do jy = max(j-10,20), min(j+10,nydim)
         ky = ky + 1
         i4med(ky) = i4buf(jy,k)
       end do
       call iindexx(ky,i4med,indx)
       i4buff(j,k) = i4med(indx((ky+1)/2))
     end do
   end do

   spcfrc=0.
   do i = minspec, maxspec
     do ky = 1, nydim
       h1 = noisemap(0,i-minspec,ky-1)
       h2 = real(max(i4buff(ky,i),1)) / expo

       spcfrc(ky,i) = amin1(amax1(h1/h2,0.0),noiseparam(0))
       spcfrc(ky,i) = spcfrc(ky,i) * noiseparam(iccd)

!       if (iccd.eq.4) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 1.04
!       else if (iccd.eq.5) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 1.07
!       else if (iccd.eq.6) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 0.95
!       else if (iccd.le.3) then
!         spcfrc(ky,i) = spcfrc(ky,i) * 0.98
!       endif

     enddo
   enddo

!   open(1,file='spcfrc.dat',status='unknown')
!   do j = 1, nydim
!     write(1,'(40f5.2)') (spcfrc(j,k),k=20,maxspec)
!   end do
!   close(1)

   idum=-1
   ksel=0
   do i=0,nevent-1
     i2pha = phaCol(i)
     if (i2pha > maxspec) then
       noiseCol(i) = 0
     else
       iy = rawYCol(i)
       if (ran2(idum).gt.dble(spcfrc(iy,i2pha))) then
         noiseCol(i) = 0
       else
         noiseCol(i) = 1
         ksel = ksel + 1
       endif
     endif
   end do

   write(errstr,'(i9,a,f6.2,a)') &
                ksel,' events were noise flagged  (', &
                100.0*real(ksel)/real(nevent),' % )'
   call message(AppMsg,Sparse)
   write(errstr,'(1x,a,f12.2,a)') 'exposure: ',expo,' [s]'
   call message(AppMsg,Sparse)

   call release(noise)

   endif ! calibrated_noise

   call release(offCor)
   call release(rawX)
   call release(rawY)
   call release(pha)
   call release(flag)

! replace EVATT_TRAILING with EVATT_BELOW_ACCEPTANCE
! once EVATT_BELOW_ACCEPTANCE has been added to the
! Pn rejection mask
!   call setEventAttribute(events,EVATT_BELOW_ACCEPTANCE)
   call setEventAttribute(events,EVATT_TRAILING)

   endif ! calibrated_offset

   call release(ccdnr)

   else ! TIMING mode flare screening

!---------------------------------------------------
!---------------------------------------------------
!-  Vadims flare code start
!---------------------------------------------------
!---------------------------------------------------

   time = column(events,"TIME",READ)
   pha = column(events,"PHA",READ)

   timeCol     => real64Data(time)
   phaCol      => int16Data(pha)


   if (isubmode .ne. FAST_TIMING) &
     call warning("WrongSubMode","Wrong observing mode: should be Timing Mode")

! append FLARE column
   if (hasColumn(events,"FLARE")) then
     flare = column(events,"FLARE",MODIFY)  
   else
     flare = addColumn (events, "FLARE", Integer16, "")
   end if   
   flareCol => int16Data(flare)
   flareCol = 0


   ndlc=2000000

!  withsoftflarescreening -- optional,type=bool,default=no; turns on flare screening code
!  softflarethreshold1 -- optional,type=real,default=8; threshold for flare screening
!  softflarethreshold2 -- optional,type=real,default=1; threshold for flare screening
!  softflaresmooth -- optional,type=string,default=BOX; smoothing method for flare screening (on of BOX, GAUSS, or FLARE)
!  softflareenergyrange -- optional,type=list<real>,default='0 60'; energy range for flare screening
!  softflaresmoothparams -- optional,type=list<real>,default='5 0 0'; smoothing parameters

   write(errstr,*) 'Soft flare screening started'
   call message(AppMsg,Sparse)


!-- energy range parameter
   npharangeelements=parametercount("softflareenergyrange")
   do i=0,npharangeelements-1
     thrpha(i) = intParameter("softflareenergyrange",i)
   end do
   if (npharangeelements == 1) then 
     thrpha(1)= thrpha(0)
     thrpha(0)=0
   endif

!-- smoothing  parameters
!--   nsmoothelements=parametercount("softflaresmoothparams")
!--   do i=0,nsmoothelements-1
     smo(0) = realParameter("softflaresmoothparams",0)
!--  end do
   
!-- cutoff threshhold for boxcar smoothed lightcurve 
     thrsig = realParameter("softflarethreshold1")

!--  make low energy lightcurve below pha threshhold
!--
  !!!!! PROBLEM: tbinsize is not an integer multiple of the frame time
     tbinsize = 0.1d0  ! should be either 0.1013989 or 0.09543424
  !!!!!
     lc=0d0
     t0=timeCol(0)
     tz=timeCol(nevent-1)
     ndlc = NINT((tz-t0)/tbinsize)+3
     allocate(lc(1:ndlc),lcs(1:ndlc))
     kmax=0
     do i=0,nevent-1 
       if (phaCol(i) .le. thrpha(1) .and. phaCol(i) .ge. thrpha(0)) then
	        k=nint((timeCol(i)-t0)/tbinsize)+1
	        if (k .lt. ndlc) then 
	          lc(k)=lc(k)+1
         else
           write(errstr,'(1x,a,i8,a,f7.4,a)') &
             'Lightcurve has more bins than intrinsically assigned ',&
              ndlc,' for ',tbinsize,' [s] bins' 
	          call error("lightCurveBins")
!          call error("Lightcurve has more bins than intrinsically assigned 2000000 for 0.1sec bins Hex # F8A67B secret file 21b SAS")
	        endif
	      endif
       kmax=max(kmax,k)
     enddo

!--  smooth light curve with boxcar
!--

     sbox=nint(smo(0))
     do i = 1, kmax
       k=0
       ss=0
       do j = max(i-sbox,0), min(i+sbox,kmax)
         k=k+1
	        ss=ss+lc(j)
       enddo
       lcs(i) = ss/real(k)
     enddo

!--  set flare flag 
!--
     flareCol=0
     flagged=0
     do i=0, nevent-1
       if (phaCol(i) .le. thrpha(1) .and. phaCol(i) .ge. thrpha(0)) then
          k=nint((timeCol(i)-t0)/tbinsize)+1
          if (lcs(k) .gt. thrsig) then
            flareCol(i)=1
            flagged=flagged+1
          end if
       endif
     enddo


   write(errstr,*) 'Soft flare screening completed ', flagged, ' of', nevent ,' events flagged'
   call message(AppMsg,Sparse)

   call release(flare)

   call release(time)
   call release(pha)

   deallocate(lc,lcs)

!---------------------------------------------------
!-  Vadims flare code end
!---------------------------------------------------
   end if


   call release(events)
   call release(infile)

   if (ldebug) then 

!   write debug images:
    call message(AppMsg,Sparse,"Writing debug images ...")

!      open(1,file='spcfrc.dat',status='unknown')
!      do j = 1, nydim
!        write(1,'(40f5.2)') (spcfrc(j,k),k=20,maxspec)
!      end do
!      close(1)

      npix(1) = nxdim
      npix(2) = nydim

      write(string,'(a4,i2.2,a4)') 'mask',iccd,'.img'         
      outfile=dataset(string,Create)
      ima=addArray(outfile,"PRIMARY",Integer16,npix)
      i2map => Int16Array2Data(ima)

      do ix=0,nxdim-1
        do iy=0,nydim-1
          if (mask(ix+1,iy+1)) then
            i2map(ix,iy)=1
          else
            i2map(ix,iy)=0
          end if
        end do
      end do          

      call release(ima)
      call release(outfile)   

!      write(string,'(a6,i2.2,a4)') 'adu20_',iccd,'.img'         
!      outfile=dataset(string,Create)
!      ima=addArray(outfile,"PRIMARY",Integer16,npix)
!      i2map => Int16Array2Data(ima)
      
!      do ix=0,nxdim-1
!        do iy=0,nydim-1
!          i2map(ix,iy)=i4pmap(ix+1,iy+1)
!        end do
!      end do          

!      call release(ima)
!      call release(outfile)   

       write(string,'(a6,i2.2,a4)') 'adu30_',iccd,'.img'         
       outfile=dataset(string,Create)
       ima=addArray(outfile,"PRIMARY",Integer16,npix)
       i2map => Int16Array2Data(ima)
      
       do ix=0,nxdim-1
         do iy=0,nydim-1
           i2map(ix,iy)=i30map(ix+1,iy+1)
         end do
       end do          

       call release(ima)
       call release(outfile)   

      write(string,'(a6,i2.2,a4)') 'offcor',iccd,'.img'         
      outfile=dataset(string,Create)
      ima=addArray(outfile,"PRIMARY",Integer16,npix)
      i2map => Int16Array2Data(ima)
      
      do ix=0,nxdim-1
        do iy=0,nydim-1
          i2map(ix,iy)=i2pmap(ix+1,iy+1)
        end do
      end do          

      call release(ima)
      call release(outfile)   

!      write(string,'(a6,i2.2,a4)') 'spcmap',iccd,'.img'         
!      outfile=dataset(string,Create)
!      ima=addArray(outfile,"PRIMARY",Real32,npix)
!      r4map => real32Array2Data(ima)
!
!      do iy = 1, nydim
!        h = 0.0
!        do i = 22, 30
!          h = h + real(i4spc(iccd,iy,i))
!        end do
!        h = h / texp
!        do ix = 0, 63
!          r4map(ix,iy-1) = h
!        end do
!      end do

!      call release(ima)
!      call release(outfile)

!      write(string,'(a6,i2.2,a4)') 'bufmap',iccd,'.img'         
!      outfile=dataset(string,Create)
!      ima=addArray(outfile,"PRIMARY",Real32,npix)
!      r4map => real32Array2Data(ima)

!      do iy = 1, nydim
!        h = 0.0
!        do i = 22, 30
!          h = h + real(i4buf(iy,i))
!        end do
!        h = h / texp
!        do ix = 0, 63
!          r4map(ix,iy-1) = h
!        end do
!      end do

!      call release(ima)
!      call release(outfile)

!      write(string,'(a6,i2.2,a4)') 'frcmap',iccd,'.img'         
!      outfile=dataset(string,Create)
!      ima=addArray(outfile,"PRIMARY",Real32,npix)
!      r4map => real32Array2Data(ima)

!      do iy = 1, nydim
!        h = 0.0
!        do i = 22, 30
!          h = h + real(spcfrc(iy,i))
!        end do
!        do ix = 0, 63
!          r4map(ix,iy-1) = h
!        end do
!      end do

!      call release(ima)
!      call release(outfile)

   end if

   end subroutine epreject
 
      FUNCTION ran2(idum)
      INTEGER idum,IM1,IM2,IMM1,IA1,IA2,IQ1,IQ2,IR1,IR2,NTAB,NDIV
      DOUBLE PRECISION ran2,AM,EPS,RNMX
      PARAMETER (IM1=2147483563,IM2=2147483399,AM=1.d0/IM1,IMM1=IM1-1, &
      IA1=40014,IA2=40692,IQ1=53668,IQ2=52774,IR1=12211,IR2=3791, &
      NTAB=32,NDIV=1+IMM1/NTAB,EPS=1.2d-7,RNMX=1.d0-EPS)
      INTEGER idum2,j,k,iv(NTAB),iy
      SAVE iv,iy,idum2
      DATA idum2/123456789/, iv/NTAB*0/, iy/0/
      if (idum.le.0) then
        idum=max(-idum,1)
        idum2=idum
        do j=NTAB+8,1,-1
          k=idum/IQ1
          idum=IA1*(idum-k*IQ1)-k*IR1
          if (idum.lt.0) idum=idum+IM1
          if (j.le.NTAB) iv(j)=idum
        end do
        iy=iv(1)
      endif
      k=idum/IQ1
      idum=IA1*(idum-k*IQ1)-k*IR1
      if (idum.lt.0) idum=idum+IM1
      k=idum2/IQ2
      idum2=IA2*(idum2-k*IQ2)-k*IR2
      if (idum2.lt.0) idum2=idum2+IM2
      j=1+iy/NDIV
      iy=iv(j)-idum2
      iv(j)=idum
      if(iy.lt.1)iy=iy+IMM1
      ran2=min(AM*iy,RNMX)
      return
      END function ran2

end module epreject_mod
